#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include "dosfs.h"
#include "propeller.h"

//#define FS_DEBUG

#define MAX_FILES   4
#define BASE_FD     3

typedef struct {
    FILEINFO file;
    int inUse;
} File;

static VOLINFO vinfo;
static File Files[MAX_FILES];
static char scratch[SECTOR_SIZE];
static int fileInit = FALSE;

static int file_write(int fd, char *buf, int nbytes);
static int file_read(int fd, char *buf, int nbytes);

int InitFileIO(void)
{
	file_read_ptr = file_read;
	file_write_ptr = file_write;
	return 0;
}

int MountFS(void)
{
    uint8_t active, ptype;
    uint32_t start, psize, err;
    int i;

#ifdef FS_DEBUG
	iprintf("GetPtnStart:\r\n");
#endif
	start = DFS_GetPtnStart(0, scratch, 0, &active, &ptype, &psize);
#ifdef FS_DEBUG
	iprintf(" -->%ld\r\n", start);
#endif
	if (start == DFS_ERRMISC)
        return FALSE;
#ifdef FS_DEBUG
    iprintf("active %d, ptype %d, psize %lx, start %lx\r\n", active, ptype, psize, start);
#endif

#ifdef FS_DEBUG
	iprintf("GetVolInfo:\r\n");
#endif
	err = DFS_GetVolInfo(0, scratch, start, &vinfo);
#ifdef FS_DEBUG
	iprintf(" -->%ld\r\n", err);
#endif
	if (err != DFS_OK)
        return FALSE;

#ifdef FS_DEBUG
	iprintf("Volume label '%-11.11s'\r\n", vinfo.label);
	iprintf("%d sector/s per cluster, %d reserved sector/s, volume total %ld sectors.\r\n", vinfo.secperclus, vinfo.reservedsecs, vinfo.numsecs);
	iprintf("%ld sectors per FAT, first FAT at sector #%ld, root dir at #%ld.\r\n",vinfo.secperfat,vinfo.fat1,vinfo.rootdir);
	iprintf("(For FAT32, the root dir is a CLUSTER number, FAT12/16 it is a SECTOR number)\r\n");
	iprintf("%d root dir entries, data area commences at sector #%ld.\r\n",vinfo.rootentries,vinfo.dataarea);
	iprintf("%ld clusters (%ld bytes) in data area, filesystem IDd as ", vinfo.numclusters, vinfo.numclusters * vinfo.secperclus * SECTOR_SIZE);
	if (vinfo.filesystem == FAT12)
		iprintf("FAT12.\r\n");
	else if (vinfo.filesystem == FAT16)
		iprintf("FAT16.\r\n");
	else if (vinfo.filesystem == FAT32)
		iprintf("FAT32.\r\n");
	else {
		iprintf("[unknown]\r\n");
        return FALSE;
    }
#endif

    for (i = 0; i < MAX_FILES; ++i)
        Files[i].inUse = FALSE;

	fileInit = TRUE;

    return TRUE;
}

static int file_write(int fd, char *buf, int nbytes)
{
    int i = fd - BASE_FD;
    uint32_t count;
    int err;
    if (!fileInit || i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    err = DFS_WriteFile(&Files[i].file, scratch, buf, &count, nbytes);
    return count > 0 ? count : EOF;
}

static int file_read(int fd, char *buf, int nbytes)
{
    int i = fd - BASE_FD;
    uint32_t count;
    int err;
    if (!fileInit || i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    err = DFS_ReadFile(&Files[i].file, scratch, buf, &count, nbytes);
    return count > 0 ? count : EOF;
}

int open(const char *name, int flags, ...)
{
    int dfs_mode, i;
	if (!fileInit) {
		if (!MountFS())
			return -1;
	}
    switch (flags & O_ACCMODE) {
    case O_RDONLY:
        dfs_mode = DFS_READ;
        break;
    case O_WRONLY:
        dfs_mode = DFS_WRITE;
        break;
    default:
        return -1;
    }
    for (i = 0; i < MAX_FILES; ++i)
        if (!Files[i].inUse)
            break;
    if (i >= MAX_FILES)
        return -1;
    if (DFS_OpenFile(&vinfo, (char *)name, dfs_mode, scratch, &Files[i].file) != DFS_OK)
        return -1;
    Files[i].inUse = TRUE;
    return BASE_FD + i;
}

int close(int fd)
{
    int i = fd - BASE_FD;
    if (!fileInit || i < 0 || i >= MAX_FILES)
        return -1;
    if (!Files[i].inUse)
        return -1;
    Files[i].inUse = FALSE;
    return 0;
}

off_t lseek(int fd, off_t offset, int whence)
{
    return -1;
}
