/* zogload.c - ZOG Loader Program
 *
 * Copyright (c) 2010 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "PLoadLib.h"
#include "osint.h"

#ifndef TRUE
#define TRUE    1
#define FALSE   0
#endif

/* timeouts for waiting for ACK/NAK */
#define INITIAL_TIMEOUT     10000   // 10 seconds
#define PACKET_TIMEOUT      1000    // 1 second

/* default baud rate */
#define BAUD_RATE           115200

/* file used by sd bootloader */
#define BOOT_FILE           "zogload.run"

/* packet types */
#define TYPE_FILE_WRITE     1
#define TYPE_FLASH_WRITE    2
#define TYPE_RAM_WRITE      3
#define TYPE_DATA           4
#define TYPE_EOF            5
#define TYPE_FILE_LOAD      6
#define TYPE_RUN            7

/* download modes */
#define DOWNLOAD_MODE_NONE      0
#define DOWNLOAD_MODE_HUB       1
#define DOWNLOAD_MODE_FILE      2
#define DOWNLOAD_MODE_DATA_FILE 3
#define DOWNLOAD_MODE_MEMORY    4

/* file types */
#define FILE_TYPE_NONE          0
#define FILE_TYPE_FLASH         1
#define FILE_TYPE_RAM           2
#define FILE_TYPE_UNKNOWN       3

/* packet format: SOH pkt# type length-lo length-hi hdrchk length*data crc1 crc2 */
#define HDR_SOH     0
#define HDR_PKTN    1
#define HDR_TYPE    2
#define HDR_LEN_HI  3
#define HDR_LEN_LO  4
#define HDR_CHK     5

/* packet header and crc lengths */
#define PKTHDRLEN   6
#define PKTMAXLEN   1024
#define PKTCRCLEN   2

/* maximum length of a frame */
#define FRAMELEN    (PKTHDRLEN + PKTMAXLEN + PKTCRCLEN)

/* protocol characters */
#define SOH     0x01    /* start of a packet */
#define ACK     0x06    /* positive acknowledgement */
#define NAK     0x15    /* negative acknowledgement */

#define updcrc(crc, ch) (crctab[((crc) >> 8) & 0xff] ^ ((crc) << 8) ^ (ch))

/* defaults */
#if defined(CYGWIN) || defined(WIN32)
#define DEF_PORT    "COM1"
#endif
#ifdef LINUX
#define DEF_PORT    "/dev/ttyUSB0"
#endif
#define DEF_BOARD   "c3"

/* local variables */
static char *port;
static char *board;
static int portOpen = FALSE;
static int debugMode = FALSE;
static int helperDownloaded = FALSE;
static int fileType = FILE_TYPE_NONE;

/* c3 helper programs */
extern uint8_t serial_helper_c3_array[];
extern int serial_helper_c3_size;
extern uint8_t serial_helper_c3_debug_array[];
extern int serial_helper_c3_debug_size;

/* sdram helper programs */
extern uint8_t serial_helper_sdram_array[];
extern int serial_helper_sdram_size;
extern uint8_t serial_helper_sdram_debug_array[];
extern int serial_helper_sdram_debug_size;

/* bootloader programs */
extern uint8_t sd_loader_c3_array[];
extern int sd_loader_c3_size;
extern uint8_t sd_loader_sdram_array[];
extern int sd_loader_sdram_size;
extern uint8_t flash_loader_c3_array[];
extern int flash_loader_c3_size;

/* local function prototypes */
static void Usage(void);
static void Fatal(char *fmt, ...);
static int WriteFileToSDCard(char *path, int checkType);
static int WriteBufferToSDCard(uint8_t *buf, int size, char *path);
static int WriteFileToMemory(char *file);
static FILE *OpenAndProbeFile(char *path, char *buf, int *pSize, int *pCnt, int *pType);
static int WriteFile(FILE *fp, uint8_t *buf, int cnt, int size);
static void WriteFilename(char *path);
static int DownloadBootloader(char *loader);
static void OpenPort(void);
static int SendPacket(int type, uint8_t *buf, int len);
static int WaitForAckNak(int timeout);
static char *PacketTypeName(int type);
static void TerminalMode(void);

int main(int argc, char *argv[])
{
    int downloadMode, needLoad, i;
    int terminalMode = FALSE; 
    char *path, *loader;

    /* initialize */
    path = NULL;
    debugMode = FALSE;
    downloadMode = DOWNLOAD_MODE_NONE;
    needLoad = FALSE;

    /* get the environment settings */
    if (!(port = getenv("PORT")))
        port = DEF_PORT;
    if (!(board = getenv("BOARD")))
        board = DEF_BOARD;

    /* get the arguments */
    for(i = 1; i < argc; ++i) {

        /* handle switches */
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'b':   // select a target board
                if (argv[i][2])
                    board = &argv[i][2];
                else if (++i < argc)
                    board = argv[i];
                else
                    Usage();
                break;
            case 'w':   // write file to SD card, flash, or RAM
                switch (argv[i][2]) {
                case 'h':   // write a serial helper to hub memory
                    downloadMode = DOWNLOAD_MODE_HUB;
                    break;
                case 'e':   // write a bootloader to the EEPROM
                    if (argv[i][3])
                        loader = &argv[i][3];
                    else if (++i < argc)
                        loader = argv[i];
                    else
                        Usage();
                    if (!DownloadBootloader(loader))
                        Fatal("Bootloader download failed");
                    break;
                case 'b':   // write filename to boot file
                    if (!path)
                        Usage();
                    WriteFilename(path);
                    break;
                case 'f':   // write an image to an SD card file
                    downloadMode = DOWNLOAD_MODE_FILE;
                    needLoad = TRUE;
                    break;
                case 'd':   // write data to an SD card file
                    downloadMode = DOWNLOAD_MODE_DATA_FILE;
                    needLoad = FALSE;
                    break;
                case 'm':   // write to memory (flash or RAM)
                    downloadMode = DOWNLOAD_MODE_MEMORY;
                    needLoad = FALSE;
                    break;
                default:
                    Usage();
                }
                break;
            case 'n':
                if (argv[i][2])
                    path = &argv[i][2];
                else if (++i < argc)
                    path = argv[i];
                else
                    Usage();
                fileType = FILE_TYPE_UNKNOWN;
                needLoad = TRUE;
                break;
            case 'r':   // run code
                if ((fileType == FILE_TYPE_NONE || fileType == FILE_TYPE_FLASH) && strcmp(board, "sdram") == 0)
                    Fatal("Flash not available on: %s", board);
                if (needLoad) {
                    if (path) {
                        char *file;
                        if (!(file = strrchr(path, '/')))
                            file = path;
                        else
                            ++file; // skip past the slash
                        if (!SendPacket(TYPE_FILE_LOAD, (uint8_t *)file, strlen(file) + 1))
                            Fatal("SendPacket FILE_LOAD failed");
                    }
                    else
                        Usage();
                    needLoad = FALSE;
                }
                SendPacket(TYPE_RUN, (uint8_t *)"", 0);
                break;
            case 'p':
                if(argv[i][2])
                    port = &argv[i][2];
                else if(++i < argc)
                    port = argv[i];
                else
                    Usage();
                if (isdigit((int)port[0])) {
                    static char buf[10];
#if defined(CYGWIN) || defined(WIN32)
                    sprintf(buf, "COM%d", atoi(port));
#endif
#ifdef LINUX
                    sprintf(buf, "/dev/ttyUSB%d", atoi(port));
#endif
                    port = buf;
                }
                break;
            case 'd':
                debugMode = TRUE;
                terminalMode = TRUE;
                break;
            case 't':
                terminalMode = TRUE;
                break;
            default:
                Usage();
                break;
            }
        }

        /* handle the input filename */
        else {
            path = argv[i];
            if (downloadMode == DOWNLOAD_MODE_NONE)
                downloadMode = DOWNLOAD_MODE_MEMORY;
            switch (downloadMode) {
            case DOWNLOAD_MODE_HUB:
                OpenPort();
                pload(path, port, DOWNLOAD_RUN_BINARY);
                if (WaitForAckNak(INITIAL_TIMEOUT) != ACK)
                    Fatal("failed to connect to helper program");
                helperDownloaded = TRUE;
                msleep(100);
                break;
            case DOWNLOAD_MODE_FILE:
                if (!WriteFileToSDCard(path, TRUE))
                    Fatal("can't download: %s", path);
                break;
            case DOWNLOAD_MODE_DATA_FILE:
                if (!WriteFileToSDCard(path, FALSE))
                    Fatal("can't download: %s", path);
                break;
            case DOWNLOAD_MODE_MEMORY:
                if (!WriteFileToMemory(path))
                    Fatal("can't download: %s", path);
                break;
            default:
                Fatal("bad download mode: %d", downloadMode);
            }
        }
    }

    /* no file specified */
    if (terminalMode)
        TerminalMode();

    /* close the port */
    if (portOpen)
        serial_done();

    /* exit successfully */
    return 0;
}

/* Usage - display a usage message and exit */
static void Usage(void)
{
    fprintf(stderr, "\
usage: zogload\n\
         [ -b <type> ]   select target board (c3 | c3ram | sdram)\n\
         [ -p <port> ]   serial port (default is %s)\n\
         [ -n <name> ]   set the name of the SD card file to run with -r\n\
         [ -wh ]         write helper to hub memory and start\n\
         [ -wf ]         write image to an SD card file\n\
         [ -wm ]         write image to memory (default)\n\
         [ -wb ]         write filename to SD boot file '%s'\n\
         [ -we <type> ]  write bootloader to EEPROM (sd or flash)\n\
         [ -r ]          run the most recently written code\n\
         [ -d ]          enable serial debugging\n\
         [ -t ]          enter terminal mode after optional compile/download\n\
         [ <name> ]      file to load\n\
\n\
   For example:\n\
\n\
     zogload -p %s -b %s hello.zbn -r\n\
\n\
   will first download a helper program into hub memory\n\
   using serial port %s and will then download the file 'hello.zbn' to memory.\n\
   After the load is complete it will run the program from memory.\n\
", DEF_PORT, BOOT_FILE, DEF_PORT, DEF_BOARD, DEF_PORT);
    exit(1);
}

static int WriteFileToSDCard(char *path, int checkType)
{
    uint8_t buf[PKTMAXLEN];
    int size, cnt;
    char *file;
    FILE *fp;

    if (!(file = strrchr(path, '/')))
        file = path;
    else
        ++file; // skip past the slash

    if (checkType) {
        if ((fp = OpenAndProbeFile(path, (char *)buf, &size, &cnt, &fileType)) == NULL)
            return FALSE;
        if (fileType == FILE_TYPE_FLASH && strcmp(board, "sdram") == 0)
            Fatal("Flash not available on: %s", board);
    }
    
    else {
        int dummyFileType;
        if ((fp = OpenAndProbeFile(path, (char *)buf, &size, &cnt, &dummyFileType)) == NULL)
            return FALSE;
    }

    if (!SendPacket(TYPE_FILE_WRITE, (uint8_t *)file, strlen(file) + 1))
        Fatal("SendPacket FILE_WRITE failed");

    return WriteFile(fp, buf, cnt, size);
}

static int WriteBufferToSDCard(uint8_t *buf, int size, char *path)
{
    int remaining = size;
    char *file;
    int cnt;

    if (!(file = strrchr(path, '/')))
        file = path;
    else
        ++file; // skip past the slash

    if (!SendPacket(TYPE_FILE_WRITE, (uint8_t *)file, strlen(file) + 1))
        Fatal("SendPacket FILE_WRITE failed");

    while ((cnt = remaining) > 0) {
        if (cnt > PKTMAXLEN)
            cnt = PKTMAXLEN;
        printf("%d bytes remaining             \r", remaining); fflush(stdout);
        if (!SendPacket(TYPE_DATA, buf, cnt))
            Fatal("SendPacket DATA failed\n");
        remaining -= cnt;
        buf += cnt;
    }
    printf("%d bytes sent             \n", size);

    if (!SendPacket(TYPE_EOF, (uint8_t *)"", 0))
        Fatal("SendPacket EOF failed");

    return TRUE;
}

static int WriteFileToMemory(char *path)
{
    uint8_t buf[PKTMAXLEN];
    int size, type, cnt;
    FILE *fp;

    if ((fp = OpenAndProbeFile(path, (char *)buf, &size, &cnt, &fileType)) == NULL)
        return FALSE;

    if (fileType == FILE_TYPE_FLASH)
        type = TYPE_FLASH_WRITE;
    else
        type = TYPE_RAM_WRITE;

    if (fileType == FILE_TYPE_FLASH && strcmp(board, "sdram") == 0)
        Fatal("Flash not available on: %s", board);

    if (!SendPacket(type, (uint8_t *)"", 0))
        Fatal("SendPacket %s failed", PacketTypeName(type));

    return WriteFile(fp, buf, cnt, size);
}

static FILE *OpenAndProbeFile(char *path, char *buf, int *pSize, int *pCnt, int *pType)
{
    FILE *fp;

    if ((fp = fopen(path, "rb")) == NULL)
        return NULL;

    fseek(fp, 0, SEEK_END);
    *pSize = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    if ((*pCnt = fread(buf, 1, PKTMAXLEN, fp)) < sizeof(uint32_t))
        Fatal("bad file header: %s", path);

    if (*(uint32_t *)buf == 0)
        *pType = FILE_TYPE_RAM;
    else
        *pType = FILE_TYPE_FLASH;

    return fp;
}

static int WriteFile(FILE *fp, uint8_t *buf, int cnt, int size)
{
    int remaining = size;

    while (cnt > 0) {
        printf("%d bytes remaining             \r", remaining); fflush(stdout);
        if (!SendPacket(TYPE_DATA, buf, cnt))
            Fatal("SendPacket DATA failed\n");
        remaining -= cnt;
        cnt = fread(buf, 1, PKTMAXLEN, fp);
    }
    printf("%d bytes sent             \n", size);

    fclose(fp);

    if (!SendPacket(TYPE_EOF, (uint8_t *)"", 0))
        Fatal("SendPacket EOF failed");

    return TRUE;
}

static void WriteFilename(char *path)
{
    char *file;

    if (!(file = strrchr(path, '/')))
        file = path;
    else
        ++file; // skip past the slash

    if (!WriteBufferToSDCard((uint8_t *)file, strlen(file) + 1, BOOT_FILE))
        Fatal("Writing '%s' to SD card failed", BOOT_FILE);
}

static int DownloadBootloader(char *loader)
{
    uint8_t *loader_array;
    int loader_size;

    if (strcmp(loader, "sd") == 0) {
        if (strcmp(board, "c3") == 0 || strcmp(board, "c3ram") == 0) {
            loader_array = sd_loader_c3_array;
            loader_size = sd_loader_c3_size;
        }
        else if (strcmp(board, "sdram") == 0) {
            loader_array = sd_loader_sdram_array;
            loader_size = sd_loader_sdram_size;
        }
        else
            Fatal("the 'sd' loader only works on the 'c3ram' or 'sdram' boards");
    }

    else if (strcmp(loader, "flash") == 0) {
        if (strcmp(board, "c3") == 0 || strcmp(board, "c3ram") == 0) {
            loader_array = flash_loader_c3_array;
            loader_size = flash_loader_c3_size;
        }
        else
            Fatal("the 'flash' loader only works on the 'c3' board");
    }

    else
        Usage();

    OpenPort();
    return ploadbuf(loader_array, loader_size, port, DOWNLOAD_EEPROM) == 0;
}

static const uint16_t crctab[256] = {
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

static uint8_t snd_pktn = 0;

/* OpenPort - open port if it hasn't already been done */
static void OpenPort(void)
{
    if (!portOpen) {
        if (!serial_init(port, BAUD_RATE))
            Fatal("Can't open port: %s", port);
        portOpen = TRUE;
    }
}

static void DownloadHelper(void)
{
    uint8_t *serial_helper_array;
    int serial_helper_size;

    /* handle C3 helper programs */
    if (strcmp(board, "c3") == 0 || strcmp(board, "c3ram") == 0) {
        if (debugMode) {
            serial_helper_array = serial_helper_c3_debug_array;
            serial_helper_size = serial_helper_c3_debug_size;
        }
        else {
            serial_helper_array = serial_helper_c3_array;
            serial_helper_size = serial_helper_c3_size;
        }
    }

    /* handle SDRAM helper programs */
    else if (strcmp(board, "sdram") == 0) {
        if (debugMode) {
            serial_helper_array = serial_helper_sdram_debug_array;
            serial_helper_size = serial_helper_sdram_debug_size;
        }
        else {
            serial_helper_array = serial_helper_sdram_array;
            serial_helper_size = serial_helper_sdram_size;
        }
    }

    /* unknown board type */
    else
        Usage();

    /* load the helper */
    ploadbuf(serial_helper_array, serial_helper_size, port, DOWNLOAD_RUN_BINARY);
    helperDownloaded = TRUE;
    msleep(100);
}

static int SendPacket(int type, uint8_t *buf, int len)
{
    uint8_t hdr[PKTHDRLEN], crc[PKTCRCLEN], *p;
    uint16_t crc16 = 0;
    int cnt, ch;

    /* open port if it hasn't already been done */
    OpenPort();

    /* download helper program if it hasn't already been done */
    if (!helperDownloaded)
        DownloadHelper();

    /* setup the frame header */
    hdr[HDR_SOH] = SOH;                                 /* SOH */
    hdr[HDR_PKTN] = ++snd_pktn;                         /* packet# */
    hdr[HDR_TYPE] = type;                               /* type type */
    hdr[HDR_LEN_HI] = (uint8_t)(len >> 8);              /* data length - high byte */
    hdr[HDR_LEN_LO] = (uint8_t)len;                     /* data length - low byte */
    hdr[HDR_CHK] = hdr[1] + hdr[2] + hdr[3] + hdr[4];   /* header checksum */

    /* compute the crc */
    for (p = buf, cnt = len; --cnt >= 0; ++p)
        crc16 = updcrc(crc16, *p);
    crc16 = updcrc(crc16, '\0');
    crc16 = updcrc(crc16, '\0');

    /* add the crc to the frame */
    crc[0] = (uint8_t)(crc16 >> 8);
    crc[1] = (uint8_t)crc16;

    /* send the packet */
    tx(hdr, PKTHDRLEN);
    if (len > 0)
        tx(buf, len);
    tx(crc, PKTCRCLEN);

    /* wait for an ACK/NAK */
    if ((ch = WaitForAckNak(PACKET_TIMEOUT)) < 0) {
        printf("Timeout waiting for ACK/NAK\n");
        ch = NAK;
    }

    /* return status */
    return ch == ACK;
}

static int WaitForAckNak(int timeout)
{
    uint8_t buf[1];
    return rx_timeout(buf, 1, timeout) == 1 ? buf[0] : -1;
}

#if 0
static int ReceivePacket(int *pType, uint8_t *buf, int len)
{
    uint8_t hdr[PKTHDRLEN], crc[PKTCRCLEN];
    int actual_len, chk;
    uint16_t crc16 = 0;

    /* look for start of packet */
    do {
        rx(&hdr[HDR_SOH], 1);
    } while (hdr[HDR_SOH] != SOH);

    /* receive the rest of the header */
    rx(&hdr[HDR_PKTN], PKTHDRLEN - 1);

    /* check the header checksum */
	chk = (hdr[1] + hdr[2] + hdr[3] + hdr[4]) & 0xff;
    if (hdr[HDR_CHK] != chk)
        return -1;

    /* make sure the buffer is big enough for the payload */
    actual_len = hdr[HDR_LEN_HI] << 8 | hdr[HDR_LEN_LO];
    if (actual_len > len)
        return -1;
    
    /* receive the packet payload */
    rx(buf, actual_len);

    /* compute the crc */
    for (len = actual_len; --len >= 0; )
        crc16 = updcrc(crc16, *buf++);

    /* receive the crc */
    rx(crc, PKTCRCLEN);

    /* check the crc */
    crc16 = updcrc(crc16, crc[0]);
    crc16 = updcrc(crc16, crc[1]);
    if (crc16 != 0)
        return -1;

    /* return packet type and the length of the payload */
    *pType = hdr[HDR_TYPE];
    return actual_len;
}
#endif

static char *PacketTypeName(int type)
{
    char *typeName = "<unknown>";
    switch (type) {
    case TYPE_FILE_WRITE:   typeName = "FILE_WRITE";    break;
    case TYPE_FLASH_WRITE:  typeName = "FLASH_WRITE";   break;
    case TYPE_RAM_WRITE:    typeName = "RAM_WRITE";     break;
    case TYPE_DATA:         typeName = "DATA";          break;
    case TYPE_EOF:          typeName = "EOF";           break;
    case TYPE_FILE_LOAD:    typeName = "FILE_LOAD";     break;
    case TYPE_RUN:          typeName = "RUN";           break;
    }
    return typeName;
}

/* TerminalMode - enter terminal mode */
static void TerminalMode(void)
{
    printf("Entering terminal mode...\n");
    OpenPort();
    for (;;) {
        uint8_t buf[1];
        if (rx_timeout(buf, 1, 0) != SERIAL_TIMEOUT)
            console_putch(buf[0]);
        else if (console_kbhit()) {
            buf[0] = console_getch();
            tx(buf, 1);
        }
    }
}

/* Fatal - display a fatal error message and exit */
static void Fatal(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fprintf(stderr, "error: ");
    vfprintf(stderr, fmt, ap);
    putc('\n', stderr);
    va_end(ap);
    exit(1);
}
